@@ -1,4 +1,5 @@
 import logging
+import re
 import ssl
 from dataclasses import dataclass
 from functools import wraps
@@ -37,7 +38,7 @@ class TLSAttribute(TypedAttributeSet):
     tls_version: str = typed_attribute()
 
 
-@dataclass
+@dataclass(eq=False)
 class TLSStream(ByteStream):
     """
     A stream wrapper that encrypts all sent data and decrypts received data.
@@ -157,10 +158,12 @@ class TLSStream(ByteStream):
 
     async def send_eof(self) -> None:
         tls_version = self.extra(TLSAttribute.tls_version)
-        version_tuple = tuple(int(part) for part in tls_version.split('.'))
-        if version_tuple < (1, 3):
-            raise NotImplementedError(f'send_eof() requires at least TLS v1.3; current session '
-                                      f'uses {tls_version}')
+        match = re.match(r'TLSv(\d+)(?:\.(\d+))?', tls_version)
+        if match:
+            major, minor = int(match.group(1)), int(match.group(2) or 0)
+            if (major, minor) < (1, 3):
+                raise NotImplementedError(f'send_eof() requires at least TLSv1.3; current '
+                                          f'session uses {tls_version}')
 
         raise NotImplementedError('send_eof() has not yet been implemented for TLS streams')
 
@@ -181,7 +184,7 @@ class TLSStream(ByteStream):
         }
 
 
-@dataclass
+@dataclass(eq=False)
 class TLSListener(Listener[TLSStream]):
     """
     A convenience listener that wraps another listener and auto-negotiates a TLS session on every
@@ -236,7 +239,7 @@ class TLSListener(Listener[TLSStream]):
         async def handler_wrapper(stream: AnyByteStream):
             from .. import fail_after
             try:
-                async with fail_after(self.handshake_timeout):
+                with fail_after(self.handshake_timeout):
                     wrapped_stream = await TLSStream.wrap(
                         stream, ssl_context=self.ssl_context,
                         standard_compatible=self.standard_compatible)
