@@ -1,14 +1,20 @@
 import asyncio
+import re
 import sys
+import time
 
-import curio
 import pytest
 import trio
 
 import anyio
 from anyio import (
-    ExceptionGroup, create_task_group, current_effective_deadline, current_time, fail_after,
-    move_on_after, open_cancel_scope, sleep, wait_all_tasks_blocked)
+    CancelScope, ExceptionGroup, create_task_group, current_effective_deadline, current_time,
+    fail_after, get_cancelled_exc_class, move_on_after, sleep, wait_all_tasks_blocked)
+
+if sys.version_info < (3, 7):
+    current_task = asyncio.Task.current_task
+else:
+    current_task = asyncio.current_task
 
 pytestmark = pytest.mark.anyio
 
@@ -26,9 +32,9 @@ async def test_already_closed():
         pass
 
     with pytest.raises(RuntimeError) as exc:
-        await tg.spawn(async_error, 'fail')
+        tg.start_soon(async_error, 'fail')
 
-    exc.match('This task group is not active; no new tasks can be spawned')
+    exc.match('This task group is not active; no new tasks can be started')
 
 
 async def test_success():
@@ -37,49 +43,171 @@ async def test_success():
 
     results = set()
     async with create_task_group() as tg:
-        await tg.spawn(async_add, 'a')
-        await tg.spawn(async_add, 'b')
+        tg.start_soon(async_add, 'a')
+        tg.start_soon(async_add, 'b')
 
     assert results == {'a', 'b'}
 
 
 @pytest.mark.parametrize('module', [
     pytest.param(asyncio, id='asyncio'),
-    pytest.param(curio, id='curio'),
     pytest.param(trio, id='trio')
 ])
 def test_run_natively(module):
     async def testfunc():
         async with create_task_group() as tg:
-            await tg.spawn(sleep, 0)
+            tg.start_soon(sleep, 0)
 
     if module is asyncio:
         from anyio._backends._asyncio import native_run
         try:
             native_run(testfunc())
         finally:
-            asyncio.set_event_loop(asyncio.new_event_loop())
+            asyncio.set_event_loop(None)
     else:
         module.run(testfunc)
 
 
-async def test_spawn_while_running():
+async def test_start_soon_while_running():
     async def task_func():
-        await tg.spawn(sleep, 0)
+        tg.start_soon(sleep, 0)
 
     async with create_task_group() as tg:
-        await tg.spawn(task_func)
+        tg.start_soon(task_func)
 
 
-async def test_spawn_after_error():
+async def test_start_soon_after_error():
     with pytest.raises(ZeroDivisionError):
         async with create_task_group() as tg:
             a = 1 / 0  # noqa: F841
 
     with pytest.raises(RuntimeError) as exc:
-        await tg.spawn(sleep, 0)
+        tg.start_soon(sleep, 0)
 
-    exc.match('This task group is not active; no new tasks can be spawned')
+    exc.match('This task group is not active; no new tasks can be started')
+
+
+async def test_start_no_value():
+    async def taskfunc(*, task_status):
+        task_status.started()
+
+    async with create_task_group() as tg:
+        value = await tg.start(taskfunc)
+        assert value is None
+
+
+async def test_start_with_value():
+    async def taskfunc(*, task_status):
+        task_status.started('foo')
+
+    async with create_task_group() as tg:
+        value = await tg.start(taskfunc)
+        assert value == 'foo'
+
+
+async def test_start_crash_before_started_call():
+    async def taskfunc(*, task_status):
+        raise Exception('foo')
+
+    async with create_task_group() as tg:
+        with pytest.raises(Exception) as exc:
+            await tg.start(taskfunc)
+
+    exc.match('foo')
+
+
+async def test_start_crash_after_started_call():
+    async def taskfunc(*, task_status):
+        task_status.started(2)
+        raise Exception('foo')
+
+    with pytest.raises(Exception) as exc:
+        async with create_task_group() as tg:
+            value = await tg.start(taskfunc)
+
+    exc.match('foo')
+    assert value == 2
+
+
+async def test_start_no_started_call():
+    async def taskfunc(*, task_status):
+        pass
+
+    async with create_task_group() as tg:
+        with pytest.raises(RuntimeError) as exc:
+            await tg.start(taskfunc)
+
+    exc.match('hild exited')
+
+
+async def test_start_cancelled():
+    async def taskfunc(*, task_status):
+        nonlocal started, finished
+        started = True
+        await sleep(2)
+        finished = True
+
+    started = finished = False
+    async with create_task_group() as tg:
+        tg.cancel_scope.cancel()
+        await tg.start(taskfunc)
+
+    assert started
+    assert not finished
+
+
+@pytest.mark.parametrize('anyio_backend', ['asyncio'])
+async def test_start_native_host_cancelled():
+    async def taskfunc(*, task_status):
+        nonlocal started, finished
+        started = True
+        await sleep(2)
+        finished = True
+
+    async def start_another():
+        async with create_task_group() as tg:
+            await tg.start(taskfunc)
+
+    started = finished = False
+    task = asyncio.get_event_loop().create_task(start_another())
+    await wait_all_tasks_blocked()
+    task.cancel()
+    with pytest.raises(asyncio.CancelledError):
+        await task
+
+    assert started
+    assert not finished
+
+
+@pytest.mark.parametrize('anyio_backend', ['asyncio'])
+async def test_start_native_child_cancelled():
+    async def taskfunc(*, task_status):
+        nonlocal task, finished
+        task = current_task()
+        await sleep(2)
+        finished = True
+
+    async def start_another():
+        async with create_task_group() as tg2:
+            await tg2.start(taskfunc)
+
+    task = None
+    finished = False
+    async with create_task_group() as tg:
+        tg.start_soon(start_another)
+        await wait_all_tasks_blocked()
+        task.cancel()
+
+    assert not finished
+
+
+async def test_start_exception_delivery():
+    def task_fn(*, task_status):
+        task_status.started("hello")
+
+    async with anyio.create_task_group() as tg:
+        with pytest.raises(TypeError, match='to be synchronous$'):
+            await tg.start(task_fn)
 
 
 async def test_host_exception():
@@ -91,7 +219,7 @@ async def test_host_exception():
     result = None
     with pytest.raises(Exception) as exc:
         async with create_task_group() as tg:
-            await tg.spawn(set_result, 'a')
+            tg.start_soon(set_result, 'a')
             raise Exception('dummy error')
 
     exc.match('dummy error')
@@ -109,9 +237,9 @@ async def test_edge_cancellation():
 
     marker = None
     async with create_task_group() as tg:
-        await tg.spawn(dummy)
+        tg.start_soon(dummy)
         assert marker is None
-        await tg.cancel_scope.cancel()
+        tg.cancel_scope.cancel()
 
     assert marker == 1
 
@@ -124,7 +252,7 @@ async def test_failing_child_task_cancels_host():
     sleep_completed = False
     with pytest.raises(Exception) as exc:
         async with create_task_group() as tg:
-            await tg.spawn(child)
+            tg.start_soon(child)
             await sleep(0.5)
             sleep_completed = True
 
@@ -141,7 +269,7 @@ async def test_failing_host_task_cancels_children():
     sleep_completed = False
     with pytest.raises(Exception) as exc:
         async with create_task_group() as tg:
-            await tg.spawn(child)
+            tg.start_soon(child)
             await wait_all_tasks_blocked()
             raise Exception('foo')
 
@@ -152,18 +280,18 @@ async def test_failing_host_task_cancels_children():
 async def test_cancel_scope_in_another_task():
     async def child():
         nonlocal result, local_scope
-        async with open_cancel_scope() as local_scope:
+        with CancelScope() as local_scope:
             await sleep(2)
             result = True
 
     local_scope = None
     result = False
     async with create_task_group() as tg:
-        await tg.spawn(child)
+        tg.start_soon(child)
         while local_scope is None:
             await sleep(0)
 
-        await local_scope.cancel()
+        local_scope.cancel()
 
     assert not result
 
@@ -175,25 +303,21 @@ async def test_cancel_propagation():
 
         assert False
 
-    async with create_task_group() as group:
-        await group.spawn(g)
+    async with create_task_group() as tg:
+        tg.start_soon(g)
         await sleep(0)
-        await group.cancel_scope.cancel()
+        tg.cancel_scope.cancel()
 
 
 async def test_cancel_twice():
-    """
-    Test that the same task can receive two cancellations. This is mostly to test the workaround
-    for curio's philosophy that a task should not be cancelled twice.
-
-    """
+    """Test that the same task can receive two cancellations."""
     async def cancel_group():
         await wait_all_tasks_blocked()
-        await group.cancel_scope.cancel()
+        tg.cancel_scope.cancel()
 
     for _ in range(2):
-        async with create_task_group() as group:
-            await group.spawn(cancel_group)
+        async with create_task_group() as tg:
+            tg.start_soon(cancel_group)
             await sleep(1)
             pytest.fail('Execution should not reach this point')
 
@@ -213,13 +337,13 @@ async def test_cancel_exiting_task_group():
 
     async def subgroup():
         async with create_task_group() as tg2:
-            await tg2.spawn(waiter)
+            tg2.start_soon(waiter)
 
     cancel_received = False
     async with create_task_group() as tg:
-        await tg.spawn(subgroup)
+        tg.start_soon(subgroup)
         await wait_all_tasks_blocked()
-        await tg.cancel_scope.cancel()
+        tg.cancel_scope.cancel()
 
     assert cancel_received
 
@@ -227,40 +351,39 @@ async def test_cancel_exiting_task_group():
 async def test_exception_group_children():
     with pytest.raises(ExceptionGroup) as exc:
         async with create_task_group() as tg:
-            await tg.spawn(async_error, 'task1')
-            await tg.spawn(async_error, 'task2', 0.15)
+            tg.start_soon(async_error, 'task1')
+            tg.start_soon(async_error, 'task2', 0.15)
 
     assert len(exc.value.exceptions) == 2
     assert sorted(str(e) for e in exc.value.exceptions) == ['task1', 'task2']
     assert exc.match('^2 exceptions were raised in the task group:\n')
     assert exc.match(r'Exception: task\d\n----')
-    if sys.version_info < (3, 7):
-        assert repr(exc.value) == "<ExceptionGroup: Exception('task1',), Exception('task2',)>"
-    else:
-        assert repr(exc.value) == "<ExceptionGroup: Exception('task1'), Exception('task2')>"
+    assert re.fullmatch(
+        r"<ExceptionGroup: Exception\('task[12]',?\), Exception\('task[12]',?\)>",
+        repr(exc.value))
 
 
 async def test_exception_group_host():
     with pytest.raises(ExceptionGroup) as exc:
         async with create_task_group() as tg:
-            await tg.spawn(async_error, 'child', 2)
+            tg.start_soon(async_error, 'child', 2)
             await wait_all_tasks_blocked()
             raise Exception('host')
 
     assert len(exc.value.exceptions) == 2
-    assert [str(e) for e in exc.value.exceptions] == ['host', 'child']
+    assert sorted(str(e) for e in exc.value.exceptions) == ['child', 'host']
     assert exc.match('^2 exceptions were raised in the task group:\n')
     assert exc.match(r'Exception: host\n----')
 
 
 async def test_escaping_cancelled_exception():
     async with create_task_group() as tg:
-        await tg.cancel_scope.cancel()
+        tg.cancel_scope.cancel()
         await sleep(0)
 
 
 async def test_cancel_scope_cleared():
-    async with move_on_after(0.1):
+    with move_on_after(0.1):
         await sleep(1)
 
     await sleep(0)
@@ -269,14 +392,14 @@ async def test_cancel_scope_cleared():
 @pytest.mark.parametrize('delay', [0, 0.1], ids=['instant', 'delayed'])
 async def test_fail_after(delay):
     with pytest.raises(TimeoutError):
-        async with fail_after(delay) as scope:
+        with fail_after(delay) as scope:
             await sleep(1)
 
     assert scope.cancel_called
 
 
 async def test_fail_after_no_timeout():
-    async with fail_after(None) as scope:
+    with fail_after(None) as scope:
         assert scope.deadline == float('inf')
         await sleep(0.1)
 
@@ -284,14 +407,14 @@ async def test_fail_after_no_timeout():
 
 
 async def test_fail_after_after_cancellation():
-    event = anyio.create_event()
+    event = anyio.Event()
     async with anyio.create_task_group() as tg:
-        await tg.spawn(tg.cancel_scope.cancel)
+        tg.cancel_scope.cancel()
         await event.wait()
 
     block_complete = False
     with pytest.raises(TimeoutError):
-        async with anyio.fail_after(0.1):
+        with fail_after(0.1):
             await anyio.sleep(0.5)
             block_complete = True
 
@@ -301,7 +424,7 @@ async def test_fail_after_after_cancellation():
 @pytest.mark.parametrize('delay', [0, 0.1], ids=['instant', 'delayed'])
 async def test_move_on_after(delay):
     result = False
-    async with move_on_after(delay) as scope:
+    with move_on_after(delay) as scope:
         await sleep(1)
         result = True
 
@@ -311,7 +434,7 @@ async def test_move_on_after(delay):
 
 async def test_move_on_after_no_timeout():
     result = False
-    async with move_on_after(None) as scope:
+    with move_on_after(None) as scope:
         assert scope.deadline == float('inf')
         await sleep(0.1)
         result = True
@@ -322,10 +445,10 @@ async def test_move_on_after_no_timeout():
 
 async def test_nested_move_on_after():
     sleep_completed = inner_scope_completed = False
-    async with move_on_after(0.1) as outer_scope:
-        assert await current_effective_deadline() == outer_scope.deadline
-        async with move_on_after(1) as inner_scope:
-            assert await current_effective_deadline() == outer_scope.deadline
+    with move_on_after(0.1) as outer_scope:
+        assert current_effective_deadline() == outer_scope.deadline
+        with move_on_after(1) as inner_scope:
+            assert current_effective_deadline() == outer_scope.deadline
             await sleep(2)
             sleep_completed = True
 
@@ -340,12 +463,12 @@ async def test_nested_move_on_after():
 async def test_shielding():
     async def cancel_when_ready():
         await wait_all_tasks_blocked()
-        await tg.cancel_scope.cancel()
+        tg.cancel_scope.cancel()
 
     inner_sleep_completed = outer_sleep_completed = False
     async with create_task_group() as tg:
-        await tg.spawn(cancel_when_ready)
-        async with move_on_after(10, shield=True) as inner_scope:
+        tg.start_soon(cancel_when_ready)
+        with move_on_after(10, shield=True) as inner_scope:
             assert inner_scope.shield
             await sleep(0.1)
             inner_sleep_completed = True
@@ -359,15 +482,60 @@ async def test_shielding():
     assert not inner_scope.cancel_called
 
 
+async def test_cancel_from_shielded_scope():
+    async with create_task_group() as tg:
+        with CancelScope(shield=True) as inner_scope:
+            assert inner_scope.shield
+            tg.cancel_scope.cancel()
+
+        with pytest.raises(get_cancelled_exc_class()):
+            await sleep(0.01)
+
+        with pytest.raises(get_cancelled_exc_class()):
+            await sleep(0.01)
+
+
+@pytest.mark.parametrize('anyio_backend', ['asyncio'])
+async def test_cancel_host_asyncgen():
+    async def host_task():
+        nonlocal done
+        async with create_task_group() as tg:
+            with CancelScope(shield=True) as inner_scope:
+                assert inner_scope.shield
+                tg.cancel_scope.cancel()
+
+            with pytest.raises(get_cancelled_exc_class()):
+                await sleep(0)
+
+            with pytest.raises(get_cancelled_exc_class()):
+                await sleep(0)
+
+            done = True
+
+    async def host_agen_fn():
+        await host_task()
+        yield
+        pytest.fail("host_agen_fn should only be __anext__ed once")
+
+    done = False
+    host_agen = host_agen_fn()
+    try:
+        await asyncio.get_event_loop().create_task(host_agen.__anext__())
+    finally:
+        await host_agen.aclose()
+
+    assert done
+
+
 async def test_shielding_immediate_scope_cancelled():
     async def cancel_when_ready():
         await wait_all_tasks_blocked()
-        await scope.cancel()
+        scope.cancel()
 
     sleep_completed = False
     async with create_task_group() as tg:
-        async with open_cancel_scope(shield=True) as scope:
-            await tg.spawn(cancel_when_ready)
+        with CancelScope(shield=True) as scope:
+            tg.start_soon(cancel_when_ready)
             await sleep(0.5)
             sleep_completed = True
 
@@ -377,15 +545,15 @@ async def test_shielding_immediate_scope_cancelled():
 async def test_cancel_scope_in_child_task():
     async def child():
         nonlocal child_scope
-        async with open_cancel_scope() as child_scope:
+        with CancelScope() as child_scope:
             await sleep(2)
 
     child_scope = None
     host_done = False
     async with create_task_group() as tg:
-        await tg.spawn(child)
+        tg.start_soon(child)
         await wait_all_tasks_blocked()
-        await child_scope.cancel()
+        child_scope.cancel()
         await sleep(0.1)
         host_done = True
 
@@ -405,9 +573,9 @@ async def test_exception_cancels_siblings():
     sleep_completed = False
     with pytest.raises(Exception) as exc:
         async with create_task_group() as tg:
-            await tg.spawn(child, False)
+            tg.start_soon(child, False)
             await wait_all_tasks_blocked()
-            await tg.spawn(child, True)
+            tg.start_soon(child, True)
 
     exc.match('foo')
     assert not sleep_completed
@@ -415,64 +583,70 @@ async def test_exception_cancels_siblings():
 
 async def test_cancel_cascade():
     async def do_something():
-        async with create_task_group() as task_group:
-            await task_group.spawn(sleep, 1)
+        async with create_task_group() as tg2:
+            tg2.start_soon(sleep, 1)
 
         raise Exception('foo')
 
     async with create_task_group() as tg:
-        await tg.spawn(do_something)
+        tg.start_soon(do_something)
         await wait_all_tasks_blocked()
-        await tg.cancel_scope.cancel()
+        tg.cancel_scope.cancel()
 
 
 async def test_cancelled_parent():
     async def child():
-        async with open_cancel_scope():
+        with CancelScope():
             await sleep(1)
 
         raise Exception('foo')
 
     async def parent(tg):
         await wait_all_tasks_blocked()
-        await tg.spawn(child)
+        tg.start_soon(child)
 
     async with create_task_group() as tg:
-        await tg.spawn(parent, tg)
-        await tg.cancel_scope.cancel()
+        tg.start_soon(parent, tg)
+        tg.cancel_scope.cancel()
 
 
 async def test_shielded_deadline():
-    async with move_on_after(10):
-        async with open_cancel_scope(shield=True):
-            async with move_on_after(1000):
-                assert await current_effective_deadline() - await current_time() > 900
+    with move_on_after(10):
+        with CancelScope(shield=True):
+            with move_on_after(1000):
+                assert current_effective_deadline() - current_time() > 900
 
 
 async def test_deadline_reached_on_start():
-    async with move_on_after(0):
+    with move_on_after(0):
         await sleep(0)
         pytest.fail('Execution should not reach this point')
 
 
+async def test_deadline_moved():
+    with fail_after(0.1) as scope:
+        scope.deadline += 0.3
+        await sleep(0.2)
+
+
 async def test_timeout_error_with_multiple_cancellations():
     with pytest.raises(TimeoutError):
-        async with fail_after(0.1):
+        with fail_after(0.1):
             async with create_task_group() as tg:
-                await tg.spawn(sleep, 2)
+                tg.start_soon(sleep, 2)
                 await sleep(2)
 
 
 async def test_nested_fail_after():
     async def killer(scope):
         await wait_all_tasks_blocked()
-        await scope.cancel()
+        scope.cancel()
 
     async with create_task_group() as tg:
-        async with open_cancel_scope() as scope:
-            async with open_cancel_scope():
-                await tg.spawn(killer, scope)
-                async with fail_after(1):
+        with CancelScope() as scope:
+            with CancelScope():
+                tg.start_soon(killer, scope)
+                with fail_after(1):
                     await sleep(2)
                     pytest.fail('Execution should not reach this point')
 
@@ -486,14 +660,14 @@ async def test_nested_fail_after():
 async def test_nested_shield():
     async def killer(scope):
         await wait_all_tasks_blocked()
-        await scope.cancel()
+        scope.cancel()
 
     with pytest.raises(TimeoutError):
         async with create_task_group() as tg:
-            async with open_cancel_scope() as scope:
-                async with open_cancel_scope(shield=True):
-                    await tg.spawn(killer, scope)
-                    async with fail_after(0.2):
+            with CancelScope() as scope:
+                with CancelScope(shield=True):
+                    tg.start_soon(killer, scope)
+                    with fail_after(0.2):
                         await sleep(2)
 
 
@@ -518,10 +692,10 @@ async def test_exception_group_filtering():
             raise Exception('%s task failed' % name)
 
     async def fn():
-        async with anyio.create_task_group() as task_group:
-            await task_group.spawn(fail, 'parent')
-            async with anyio.create_task_group() as task_group2:
-                await task_group2.spawn(fail, 'child')
+        async with anyio.create_task_group() as tg:
+            tg.start_soon(fail, 'parent')
+            async with anyio.create_task_group() as tg2:
+                tg2.start_soon(fail, 'child')
                 await anyio.sleep(1)
 
     with pytest.raises(ExceptionGroup) as exc:
@@ -534,25 +708,25 @@ async def test_exception_group_filtering():
 
 async def test_cancel_propagation_with_inner_spawn():
     async def g():
-        async with anyio.create_task_group() as g:
-            await g.spawn(anyio.sleep, 10)
-            await anyio.sleep(5)
+        async with anyio.create_task_group() as tg2:
+            tg2.start_soon(anyio.sleep, 10)
+            await anyio.sleep(1)
 
         assert False
 
-    async with anyio.create_task_group() as group:
-        await group.spawn(g)
-        await anyio.sleep(0.1)
-        await group.cancel_scope.cancel()
+    async with anyio.create_task_group() as tg:
+        tg.start_soon(g)
+        await wait_all_tasks_blocked()
+        tg.cancel_scope.cancel()
 
 
 async def test_escaping_cancelled_error_from_cancelled_task():
     """Regression test for issue #88. No CancelledError should escape the outer scope."""
-    async with open_cancel_scope() as scope:
-        async with move_on_after(0.1):
+    with CancelScope() as scope:
+        with move_on_after(0.1):
             await sleep(1)
 
-        await scope.cancel()
+        scope.cancel()
 
 
 @pytest.mark.filterwarnings('ignore:"@coroutine" decorator is deprecated:DeprecationWarning')
@@ -560,8 +734,8 @@ def test_cancel_generator_based_task():
     from asyncio import coroutine
 
     async def native_coro_part():
-        async with open_cancel_scope() as scope:
-            await scope.cancel()
+        with CancelScope() as scope:
+            scope.cancel()
 
     @coroutine
     def generator_part():
@@ -578,9 +752,135 @@ async def test_suppress_exception_context():
     """
     with pytest.raises(ValueError) as exc:
         async with create_task_group() as tg:
-            await tg.cancel_scope.cancel()
-            async with create_task_group() as inner:
-                await inner.spawn(sleep, 1)
+            tg.cancel_scope.cancel()
+            async with create_task_group() as tg2:
+                tg2.start_soon(sleep, 1)
                 raise ValueError
 
     assert exc.value.__context__ is None
+
+
+@pytest.mark.parametrize('anyio_backend', ['asyncio'])
+async def test_cancel_native_future_tasks():
+    async def wait_native_future():
+        loop = asyncio.get_event_loop()
+        await loop.create_future()
+
+    async with anyio.create_task_group() as tg:
+        tg.start_soon(wait_native_future)
+        tg.cancel_scope.cancel()
+
+
+@pytest.mark.parametrize('anyio_backend', ['asyncio'])
+async def test_cancel_native_future_tasks_cancel_scope():
+    async def wait_native_future():
+        with anyio.CancelScope():
+            loop = asyncio.get_event_loop()
+            await loop.create_future()
+
+    async with anyio.create_task_group() as tg:
+        tg.start_soon(wait_native_future)
+        tg.cancel_scope.cancel()
+
+
+@pytest.mark.parametrize('anyio_backend', ['asyncio'])
+async def test_cancel_completed_task():
+    loop = asyncio.get_event_loop()
+    old_exception_handler = loop.get_exception_handler()
+    exceptions = []
+
+    def exception_handler(*args, **kwargs):
+        exceptions.append((args, kwargs))
+
+    loop.set_exception_handler(exception_handler)
+    try:
+        async def noop():
+            pass
+
+        async with anyio.create_task_group() as tg:
+            tg.start_soon(noop)
+            tg.cancel_scope.cancel()
+
+        assert exceptions == []
+    finally:
+        loop.set_exception_handler(old_exception_handler)
+
+
+async def test_task_in_sync_spawn_callback():
+    outer_task_id = anyio.get_current_task().id
+    inner_task_id = None
+
+    def task_wrap():
+        assert anyio.get_current_task().id == outer_task_id
+
+        async def corofn():
+            nonlocal inner_task_id
+            inner_task_id = anyio.get_current_task().id
+
+        return corofn()
+
+    async with create_task_group() as tg:
+        tg.start_soon(task_wrap)
+
+    assert inner_task_id is not None
+    assert inner_task_id != outer_task_id
+
+
+async def test_shielded_cancel_sleep_time():
+    """Test that cancelling a shielded tasks spends more time sleeping than cancelling."""
+    event = anyio.Event()
+    hang_time = 0.2
+
+    async def set_event():
+        await sleep(hang_time)
+        event.set()
+
+    async def never_cancel_task():
+        with CancelScope(shield=True):
+            await sleep(0.2)
+            await event.wait()
+
+    async with create_task_group() as tg:
+        tg.start_soon(set_event)
+
+        async with create_task_group() as tg:
+            tg.start_soon(never_cancel_task)
+            tg.cancel_scope.cancel()
+            process_time = time.process_time()
+
+        assert (time.process_time() - process_time) < hang_time
+
+
+async def test_cancelscope_wrong_exit_order():
+    """
+    Test that a RuntimeError is raised if the task tries to exit cancel scopes in the wrong order.
+
+    """
+    scope1 = CancelScope()
+    scope2 = CancelScope()
+    scope1.__enter__()
+    scope2.__enter__()
+    pytest.raises(RuntimeError, scope1.__exit__, None, None, None)
+
+
+async def test_cancelscope_exit_before_enter():
+    """Test that a RuntimeError is raised if one tries to exit a cancel scope before entering."""
+    scope = CancelScope()
+    pytest.raises(RuntimeError, scope.__exit__, None, None, None)
+
+
+@pytest.mark.parametrize('anyio_backend', ['asyncio'])  # trio does not check for this yet
+async def test_cancelscope_exit_in_wrong_task():
+    async def enter_scope(scope):
+        scope.__enter__()
+
+    async def exit_scope(scope):
+        scope.__exit__(None, None, None)
+
+    scope = CancelScope()
+    async with create_task_group() as tg:
+        tg.start_soon(enter_scope, scope)
+
+    with pytest.raises(RuntimeError):
+        async with create_task_group() as tg:
+            tg.start_soon(exit_scope, scope)
